@@ -24,6 +24,7 @@ import java.io.InputStream;
 import java.net.URL;
 import java.net.URLConnection;
 import java.net.URLStreamHandler;
+import java.security.AccessControlException;
 import java.security.AllPermission;
 import java.util.ArrayList;
 import java.util.Enumeration;
@@ -33,21 +34,19 @@ import java.util.Iterator;
 import java.util.List;
 import java.util.Map;
 import java.util.NoSuchElementException;
-import java.util.Properties;
 import java.util.Set;
 
-import org.apache.felix.framework.Felix.FelixResolver;
-import org.apache.felix.framework.ModuleImpl;
+import org.apache.felix.framework.Felix.StatefulResolver;
+import org.apache.felix.framework.capabilityset.Attribute;
+import org.apache.felix.framework.capabilityset.Capability;
+import org.apache.felix.framework.capabilityset.Directive;
+import org.apache.felix.framework.resolver.Module;
 import org.apache.felix.framework.util.FelixConstants;
 import org.apache.felix.framework.util.StringMap;
 import org.apache.felix.framework.util.Util;
-import org.apache.felix.framework.util.manifestparser.Capability;
+import org.apache.felix.framework.util.manifestparser.CapabilityImpl;
 import org.apache.felix.framework.util.manifestparser.ManifestParser;
-import org.apache.felix.framework.util.manifestparser.R4Attribute;
-import org.apache.felix.framework.util.manifestparser.R4Directive;
-import org.apache.felix.moduleloader.ICapability;
-import org.apache.felix.moduleloader.IContent;
-import org.apache.felix.moduleloader.IModule;
+import org.apache.felix.framework.resolver.Content;
 import org.osgi.framework.AdminPermission;
 import org.osgi.framework.Bundle;
 import org.osgi.framework.BundleActivator;
@@ -79,7 +78,7 @@ import org.osgi.framework.Version;
 // with the parent classloader and one instance per framework instance that
 // keeps track of extension bundles and systembundle exports for that framework
 // instance.
-class ExtensionManager extends URLStreamHandler implements IContent
+class ExtensionManager extends URLStreamHandler implements Content
 {
     // The private instance that is added to Felix.class.getClassLoader() -
     // will be null if extension bundles are not supported (i.e., we are not
@@ -118,11 +117,12 @@ class ExtensionManager extends URLStreamHandler implements IContent
 
     private final Logger m_logger;
     private final Map m_headerMap = new StringMap(false);
-    private final IModule m_systemBundleModule;
-    private ICapability[] m_capabilities = null;
+    private final Module m_systemBundleModule;
+    private List<Capability> m_capabilities = null;
     private Set m_exportNames = null;
     private Object m_securityContext = null;
     private final List m_extensions;
+    private volatile Bundle[] m_extensionsCache;
     private final Set m_names;
     private final Map m_sourceToExtensions;
 
@@ -133,6 +133,7 @@ class ExtensionManager extends URLStreamHandler implements IContent
         m_logger = null;
         m_systemBundleModule = null;
         m_extensions = new ArrayList();
+        m_extensionsCache = new Bundle[0];
         m_names = new HashSet();
         m_sourceToExtensions = new HashMap();
     }
@@ -152,6 +153,7 @@ class ExtensionManager extends URLStreamHandler implements IContent
     {
         m_systemBundleModule = new ExtensionManagerModule(felix);
         m_extensions = null;
+        m_extensionsCache = null;
         m_names = null;
         m_sourceToExtensions = null;
         m_logger = logger;
@@ -177,7 +179,10 @@ class ExtensionManager extends URLStreamHandler implements IContent
         // packages should be exported by the system bundle.
         String syspkgs = (String) felix.getConfig().get(FelixConstants.FRAMEWORK_SYSTEMPACKAGES);
         // If no system packages were specified, load our default value.
-        syspkgs = (syspkgs == null) ? loadDefaultSystemPackages(m_logger) : syspkgs;
+        syspkgs = (syspkgs == null)
+            ? Util.getDefaultProperty(logger, Constants.FRAMEWORK_SYSTEMPACKAGES)
+            : syspkgs;
+        syspkgs = (syspkgs == null) ? "" : syspkgs;
         // If any extra packages are specified, then append them.
         String extra = (String) felix.getConfig().get(FelixConstants.FRAMEWORK_SYSTEMPACKAGES_EXTRA);
         syspkgs = (extra == null) ? syspkgs : syspkgs + "," + extra;
@@ -187,12 +192,12 @@ class ExtensionManager extends URLStreamHandler implements IContent
         {
             ManifestParser mp = new ManifestParser(
                 m_logger, felix.getConfig(), m_systemBundleModule, m_headerMap);
-            ICapability[] caps = aliasSymbolicName(mp.getCapabilities());
+            List<Capability> caps = aliasSymbolicName(mp.getCapabilities());
             setCapabilities(caps);
         }
         catch (Exception ex)
         {
-            m_capabilities = new ICapability[0];
+            m_capabilities = new ArrayList<Capability>(0);
             m_logger.log(
                 Logger.LOG_ERROR,
                 "Error parsing system bundle export statement: "
@@ -200,39 +205,39 @@ class ExtensionManager extends URLStreamHandler implements IContent
         }
     }
 
-    private static ICapability[] aliasSymbolicName(ICapability[] caps)
+    private static List<Capability> aliasSymbolicName(List<Capability> caps)
     {
         if (caps == null)
         {
-            return new ICapability[0];
+            return new ArrayList<Capability>(0);
         }
 
-        ICapability[] aliasCaps = new ICapability[caps.length];
-        System.arraycopy(caps, 0, aliasCaps, 0, caps.length);
+        List<Capability> aliasCaps = new ArrayList<Capability>(caps);
 
-        for (int capIdx = 0; capIdx < aliasCaps.length; capIdx++)
+        for (int capIdx = 0; capIdx < aliasCaps.size(); capIdx++)
         {
             // Get the attributes and search for bundle symbolic name.
-            R4Attribute[] attrs = ((Capability) aliasCaps[capIdx]).getAttributes();
-            for (int i = 0; i < attrs.length; i++)
+            List<Attribute> attrs = aliasCaps.get(capIdx).getAttributes();
+            for (int i = 0; i < attrs.size(); i++)
             {
                 // If there is a bundle symbolic name attribute, add the
                 // standard alias as a value.
-                if (attrs[i].getName().equalsIgnoreCase(Constants.BUNDLE_SYMBOLICNAME_ATTRIBUTE))
+                if (attrs.get(i).getName().equalsIgnoreCase(Constants.BUNDLE_SYMBOLICNAME_ATTRIBUTE))
                 {
                     // Make a copy of the attribute array.
-                    R4Attribute[] aliasAttrs = new R4Attribute[attrs.length];
-                    System.arraycopy(attrs, 0, aliasAttrs, 0, attrs.length);
+                    List<Attribute> aliasAttrs = new ArrayList<Attribute>(attrs);
                     // Add the aliased value.
-                    aliasAttrs[i] = new R4Attribute(
+                    aliasAttrs.set(i, new Attribute(
                         Constants.BUNDLE_SYMBOLICNAME_ATTRIBUTE,
-                        new String[] { (String) attrs[i].getValue(), Constants.SYSTEM_BUNDLE_SYMBOLICNAME }, false);
+                        new String[] {
+                            (String) attrs.get(i).getValue(), Constants.SYSTEM_BUNDLE_SYMBOLICNAME},
+                        false));
                     // Create the aliased capability to replace the old capability.
-                    aliasCaps[capIdx] = new Capability(
-                        caps[capIdx].getModule(),
-                        caps[capIdx].getNamespace(),
-                        ((Capability) caps[capIdx]).getDirectives(),
-                        aliasAttrs);
+                    aliasCaps.set(capIdx, new CapabilityImpl(
+                        caps.get(capIdx).getModule(),
+                        caps.get(capIdx).getNamespace(),
+                        caps.get(capIdx).getDirectives(),
+                        aliasAttrs));
                     // Continue with the next capability.
                     break;
                 }
@@ -242,7 +247,7 @@ class ExtensionManager extends URLStreamHandler implements IContent
         return aliasCaps;
     }
 
-    public IModule getModule()
+    public Module getModule()
     {
         return m_systemBundleModule;
     }
@@ -277,8 +282,16 @@ class ExtensionManager extends URLStreamHandler implements IContent
         Object sm = System.getSecurityManager();
         if (sm != null)
         {
-            ((SecurityManager) sm).checkPermission(
-                new AdminPermission(bundle, AdminPermission.EXTENSIONLIFECYCLE));
+            try
+            {
+                ((SecurityManager) sm).checkPermission(
+                    new AdminPermission(bundle, AdminPermission.EXTENSIONLIFECYCLE));
+            }
+            catch (SecurityException ex)
+            {
+                // TODO: SECURITY - we need to throw this exception because of the 4.2.0 ct
+                throw new AccessControlException(ex.getMessage());
+            }
         }
 
         if (!((BundleProtectionDomain) bundle.getProtectionDomain()).impliesDirect(new AllPermission()))
@@ -286,7 +299,7 @@ class ExtensionManager extends URLStreamHandler implements IContent
             throw new SecurityException("Extension Bundles must have AllPermission");
         }
 
-        R4Directive dir = ManifestParser.parseExtensionBundleHeader((String)
+        Directive dir = ManifestParser.parseExtensionBundleHeader((String)
             bundle.getCurrentModule().getHeaders().get(Constants.FRAGMENT_HOST));
 
         // We only support classpath extensions (not bootclasspath).
@@ -299,15 +312,12 @@ class ExtensionManager extends URLStreamHandler implements IContent
 
         try
         {
-// TODO: EXTENSIONMANAGER - Should we be setting this?
-//            bundle.setExtension(true);
-
             // Merge the exported packages with the exported packages of the systembundle.
-            ICapability[] exports = null;
+            List<Capability> exports = null;
             try
             {
                 exports = ManifestParser.parseExportHeader(
-                    m_systemBundleModule,
+                    m_logger, m_systemBundleModule,
                     (String) bundle.getCurrentModule().getHeaders().get(Constants.EXPORT_PACKAGE),
                     m_systemBundleModule.getSymbolicName(), m_systemBundleModule.getVersion());
                 exports = aliasSymbolicName(exports);
@@ -315,6 +325,7 @@ class ExtensionManager extends URLStreamHandler implements IContent
             catch (Exception ex)
             {
                 m_logger.log(
+                    bundle,
                     Logger.LOG_ERROR,
                     "Error parsing extension bundle export statement: "
                     + bundle.getCurrentModule().getHeaders().get(Constants.EXPORT_PACKAGE), ex);
@@ -330,20 +341,18 @@ class ExtensionManager extends URLStreamHandler implements IContent
             else
             {
                 // We don't support extensions (i.e., the parent is not an URLClassLoader).
-                m_logger.log(Logger.LOG_WARNING,
+                m_logger.log(bundle, Logger.LOG_WARNING,
                     "Unable to add extension bundle to FrameworkClassLoader - Maybe not an URLClassLoader?");
                 throw new UnsupportedOperationException(
                     "Unable to add extension bundle to FrameworkClassLoader - Maybe not an URLClassLoader?");
             }
-            ICapability[] temp = new ICapability[m_capabilities.length + exports.length];
-            System.arraycopy(m_capabilities, 0, temp, 0, m_capabilities.length);
-            System.arraycopy(exports, 0, temp, m_capabilities.length, exports.length);
+            List<Capability> temp = new ArrayList<Capability>(m_capabilities.size() + exports.size());
+            temp.addAll(m_capabilities);
+            temp.addAll(exports);
             setCapabilities(temp);
         }
         catch (Exception ex)
         {
-// TODO: EXTENSIONMANAGER - Should we be setting this?
-//            bundle.setExtension(false);
             throw ex;
         }
 
@@ -367,25 +376,26 @@ class ExtensionManager extends URLStreamHandler implements IContent
         {
             try
             {
+// TODO: SECURITY - Should this consider security?
                 BundleActivator activator = (BundleActivator)
                     felix.getClass().getClassLoader().loadClass(
                         activatorClass.trim()).newInstance();
 
-// TODO: KARL - This is kind of hacky, can we improve it?
+// TODO: EXTENSIONMANAGER - This is kind of hacky, can we improve it?
                 felix.m_activatorList.add(activator);
 
-                BundleContext context = felix.getBundleContext();
+                BundleContext context = felix._getBundleContext();
 
                 bundle.setBundleContext(context);
 
-                if (felix.getState() == Bundle.ACTIVE)
+                if ((felix.getState() == Bundle.ACTIVE) || (felix.getState() == Bundle.STARTING))
                 {
                     Felix.m_secureAction.startActivator(activator, context);
                 }
             }
             catch (Throwable ex)
             {
-                m_logger.log(Logger.LOG_WARNING,
+                m_logger.log(bundle, Logger.LOG_WARNING,
                     "Unable to start Felix Extension Activator", ex);
             }
         }
@@ -406,7 +416,7 @@ class ExtensionManager extends URLStreamHandler implements IContent
         }
     }
 
-    private void setCapabilities(ICapability[] capabilities)
+    private void setCapabilities(List<Capability> capabilities)
     {
         m_capabilities = capabilities;
         m_headerMap.put(Constants.EXPORT_PACKAGE, convertCapabilitiesToHeaders(m_headerMap));
@@ -417,9 +427,9 @@ class ExtensionManager extends URLStreamHandler implements IContent
         StringBuffer exportSB = new StringBuffer("");
         Set exportNames = new HashSet();
 
-        for (int i = 0; (m_capabilities != null) && (i < m_capabilities.length); i++)
+        for (int i = 0; (m_capabilities != null) && (i < m_capabilities.size()); i++)
         {
-            if (m_capabilities[i].getNamespace().equals(ICapability.PACKAGE_NAMESPACE))
+            if (m_capabilities.get(i).getNamespace().equals(Capability.PACKAGE_NAMESPACE))
             {
                 // Add a comma separate if there is an existing package.
                 if (exportSB.length() > 0)
@@ -428,13 +438,31 @@ class ExtensionManager extends URLStreamHandler implements IContent
                 }
 
                 // Append exported package information.
-                exportSB.append(((Capability) m_capabilities[i]).getPackageName());
-                exportSB.append("; version=\"");
-                exportSB.append(((Capability) m_capabilities[i]).getPackageVersion().toString());
-                exportSB.append("\"");
+                exportSB.append(m_capabilities.get(i).getAttribute(Capability.PACKAGE_ATTR).getValue());
+                for (Directive dir : m_capabilities.get(i).getDirectives())
+                {
+                    exportSB.append("; ");
+                    exportSB.append(dir.getName());
+                    exportSB.append(":=\"");
+                    exportSB.append(dir.getValue());
+                    exportSB.append("\"");
+                }
+                for (Attribute attr : m_capabilities.get(i).getAttributes())
+                {
+                    if (!attr.getName().equals(Capability.PACKAGE_ATTR)
+                        && !attr.getName().equals(Constants.BUNDLE_SYMBOLICNAME_ATTRIBUTE)
+                        && !attr.getName().equals(Constants.BUNDLE_VERSION_ATTRIBUTE))
+                    {
+                        exportSB.append("; ");
+                        exportSB.append(attr.getName());
+                        exportSB.append("=\"");
+                        exportSB.append(attr.getValue());
+                        exportSB.append("\"");
+                    }
+                }
 
                 // Remember exported packages.
-                exportNames.add(((Capability) m_capabilities[i]).getPackageName());
+                exportNames.add(m_capabilities.get(i).getAttribute(Capability.PACKAGE_ATTR).getValue());
             }
         }
 
@@ -451,7 +479,7 @@ class ExtensionManager extends URLStreamHandler implements IContent
      * See whether any registered extension provides the class requested. If not
      * throw an IOException.
      */
-    public synchronized URLConnection openConnection(URL url) throws IOException
+    public URLConnection openConnection(URL url) throws IOException
     {
         String path = url.getPath();
 
@@ -460,17 +488,31 @@ class ExtensionManager extends URLStreamHandler implements IContent
             return new URLHandlersBundleURLConnection(url);
         }
 
-        for (Iterator iter = m_extensions.iterator(); iter.hasNext();)
+        Bundle[] extensions = m_extensionsCache;
+        URL result = null;
+        for (Bundle extBundle : extensions)
         {
-            URL result = ((BundleImpl) iter.next()).getCurrentModule().getEntry(path);
-
+            try
+            {
+                result = ((ModuleImpl) ((BundleImpl) extBundle).getCurrentModule()).getResourceLocal(path);
+            }
+            catch (Exception ex)
+            {
+                // Maybe the bundle went away, so ignore this exception.
+            }
             if (result != null)
             {
                 return result.openConnection();
             }
         }
 
-        throw new IOException("Resource not provided by any extension!");
+        return new URLConnection(url)
+            {
+                public void connect() throws IOException
+                {
+                    throw new IOException("Resource not provided by any extension!");
+                }
+            };
     }
 
     protected InetAddress getHostAddress(URL u)
@@ -492,6 +534,8 @@ class ExtensionManager extends URLStreamHandler implements IContent
         sourceExtensions.add(extension);
 
         _add(extension.getSymbolicName(), extension);
+        m_extensionsCache = (Bundle[])
+                m_extensions.toArray(new Bundle[m_extensions.size()]);
     }
 
     private synchronized void _removeExtensions(Object source)
@@ -512,6 +556,8 @@ class ExtensionManager extends URLStreamHandler implements IContent
                 Bundle bundle = (Bundle) extIter.next();
                 _add(bundle.getSymbolicName(), bundle);
             }
+            m_extensionsCache = (Bundle[])
+                m_extensions.toArray(new Bundle[m_extensions.size()]);
         }
     }
 
@@ -559,7 +605,7 @@ class ExtensionManager extends URLStreamHandler implements IContent
         return null;
     }
 
-    public IContent getEntryAsContent(String name)
+    public Content getEntryAsContent(String name)
     {
         return null;
     }
@@ -569,63 +615,21 @@ class ExtensionManager extends URLStreamHandler implements IContent
         return null;
     }
 
+    public URL getEntryAsURL(String name)
+    {
+        return null;
+    }
+
     //
     // Utility methods.
     //
 
-    /**
-     * The default name used for the default configuration properties file.
-    **/
-    public static final String DEFAULT_PROPERTIES_FILE_VALUE = "default.properties";
-
-    static String loadDefaultSystemPackages(Logger logger)
-    {
-        // If we cannot get configuration properties for any reason, then just
-        // attempt to load resource default.properties instead.
-        URL propURL = ExtensionManager.class.getClassLoader()
-            .getResource(DEFAULT_PROPERTIES_FILE_VALUE);
-        InputStream is = null;
-        try
-        {
-            // Load properties from URL.
-            is = propURL.openConnection().getInputStream();
-            Properties props = new Properties();
-            props.load(is);
-            is.close();
-            // Perform variable substitution for system properties.
-            for (Enumeration e = props.propertyNames(); e.hasMoreElements(); )
-            {
-                String name = (String) e.nextElement();
-                props.setProperty(name,
-                    Util.substVars(props.getProperty(name), name, null, props));
-            }
-            // Return system packages property.
-            return props.getProperty(Constants.FRAMEWORK_SYSTEMPACKAGES);
-        }
-        catch (Exception ex)
-        {
-            // Try to close input stream if we have one.
-            try
-            {
-                if (is != null) is.close();
-            }
-            catch (IOException ex2)
-            {
-                // Nothing we can do.
-            }
-
-            logger.log(
-                Logger.LOG_ERROR, "Unable to load any configuration properties.", ex);
-            return "";
-        }
-    }
-
     class ExtensionManagerModule extends ModuleImpl
     {
         private final Version m_version;
         ExtensionManagerModule(Felix felix)
         {
-            super(m_logger, felix, "0",
+            super(m_logger, felix.getConfig(), felix, "0",
                 felix.getBootPackages(), felix.getBootPackageWildcards());
             m_version = new Version((String)
                 felix.getConfig().get(FelixConstants.FELIX_VERSION_PROPERTY));
@@ -639,7 +643,7 @@ class ExtensionManager extends URLStreamHandler implements IContent
             }
         }
 
-        public ICapability[] getCapabilities()
+        public List<Capability> getCapabilities()
         {
             synchronized (ExtensionManager.this)
             {
@@ -659,15 +663,45 @@ class ExtensionManager extends URLStreamHandler implements IContent
 
         public Class getClassByDelegation(String name) throws ClassNotFoundException
         {
-            synchronized (ExtensionManager.this)
+            Class clazz = null;
+            String pkgName = Util.getClassPackage(name);
+            if (shouldBootDelegate(pkgName))
             {
-                if (!m_exportNames.contains(Util.getClassPackage(name)))
+                try
                 {
-                    throw new ClassNotFoundException(name);
+                    // Get the appropriate class loader for delegation.
+                    ClassLoader bdcl = getBootDelegationClassLoader();
+                    clazz = bdcl.loadClass(name);
+                    // If this is a java.* package, then always terminate the
+                    // search; otherwise, continue to look locally if not found.
+                    if (pkgName.startsWith("java.") || (clazz != null))
+                    {
+                        return clazz;
+                    }
+                }
+                catch (ClassNotFoundException ex)
+                {
+                    // If this is a java.* package, then always terminate the
+                    // search; otherwise, continue to look locally if not found.
+                    if (pkgName.startsWith("java."))
+                    {
+                        throw ex;
+                    }
                 }
             }
+            if (clazz == null)
+            {
+                synchronized (ExtensionManager.this)
+                {
+                    if (!m_exportNames.contains(Util.getClassPackage(name)))
+                    {
+                        throw new ClassNotFoundException(name);
+                    }
+                }
 
-            return getClass().getClassLoader().loadClass(name);
+                clazz = getClass().getClassLoader().loadClass(name);
+            }
+            return clazz;
         }
 
         public URL getResourceByDelegation(String name)
@@ -697,12 +731,12 @@ class ExtensionManager extends URLStreamHandler implements IContent
             return null;
         }
 
-        public FelixResolver getResolver()
+        public StatefulResolver getResolver()
         {
             return null;
         }
 
-        public void attachFragmentContents(IContent[] fragmentContents)
+        public void attachFragmentContents(Content[] fragmentContents)
             throws Exception
         {
             throw new UnsupportedOperationException("Should not be used!");
@@ -713,7 +747,7 @@ class ExtensionManager extends URLStreamHandler implements IContent
             // Nothing needed here.
         }
 
-        public IContent getContent()
+        public Content getContent()
         {
             return ExtensionManager.this;
         }
@@ -733,5 +767,10 @@ class ExtensionManager extends URLStreamHandler implements IContent
         {
             return getClass().getClassLoader().getResourceAsStream(urlPath);
         }
+
+        public URL getLocalURL(int index, String urlPath)
+        {
+            return getClass().getClassLoader().getResource(urlPath);
+        }
     }
-}
\ No newline at end of file
+}
